# coding=UTF-8
from __future__ import absolute_import, division, print_function

import collections
import io
import struct

import six

from coolamqp.framing.definitions import FRAME_HEADER, FRAME_HEARTBEAT, \
    FRAME_END, FRAME_METHOD, FRAME_BODY
from coolamqp.framing.frames import AMQPBodyFrame, AMQPHeaderFrame, \
    AMQPHeartbeatFrame, AMQPMethodFrame

FRAME_TYPES = {
    FRAME_HEADER: AMQPHeaderFrame,
    FRAME_BODY: AMQPBodyFrame,
    FRAME_METHOD: AMQPMethodFrame
}

ordpy2 = ord if six.PY2 else lambda x: x


class ReceivingFramer(object):
    """
    Assembles AMQP framing from received data.

    Just call with .put(data) upon receiving,
    and on_frame will be called with fresh frames.

    Not thread safe.

    State machine
        (frame_type is None)  and has_bytes(1)        ->
                (frame_type <- bytes(1))

        (frame_type is HEARTBEAT) and has_bytes(AMQPHeartbeatFrame.LENGTH-1)
                      ->          (output_frame, frame_type <- None)
        (frame_type is not HEARTBEAT and not None) and has_bytes(6)  ->
                                       (frame_channel <- bytes(2),
                                         frame_size <- bytes(4))

        (frame_size is not None) and has_bytes(frame_size+1)    ->
                    (output_frame,

                                    frame_type <- None
                                    frame_size < None)
    """

    def __init__(self, on_frame=lambda frame: None):
        self.chunks = collections.deque()  # all received data
        self.total_data_len = 0

        self.frame_type = None
        self.frame_channel = None
        self.frame_size = None

        self.bytes_needed = None  # bytes needed for a new frame
        self.on_frame = on_frame

    def put(self, data):
        """
        Called upon receiving data.

        May result in any number of .on_frame() calls
        :param data: received data
        """
        self.total_data_len += len(data)
        self.chunks.append(memoryview(data))

        while self._statemachine():
            pass

    def _extract_single_byte(self):
        return ordpy2(self._extract(1)[0])

    def _extract(self, up_to):
        """
        return up to up_to bytes from current chunk, switch if necessary
        """
        assert self.total_data_len >= up_to, \
            'Tried to extract %s but %s remaining' % (
                up_to, self.total_data_len)
        if up_to >= len(self.chunks[0]):
            q = self.chunks.popleft()
        else:
            q = self.chunks[0][:up_to]
            self.chunks[0] = self.chunks[0][up_to:]

        self.total_data_len -= len(q)
        assert len(q) <= up_to, 'extracted %s but %s was requested' % (
            len(q), up_to)
        return q

    def _statemachine(self):
        # state rule 1
        if self.frame_type is None and self.total_data_len > 0:
            self.frame_type = self._extract_single_byte()

            if self.frame_type not in (
                    FRAME_HEARTBEAT, FRAME_HEADER, FRAME_METHOD, FRAME_BODY):
                raise ValueError('Invalid frame')

            return True

        # state rule 2
        elif (self.frame_type == FRAME_HEARTBEAT) and (
                self.total_data_len >= AMQPHeartbeatFrame.LENGTH - 1):
            data = b''
            while len(data) < AMQPHeartbeatFrame.LENGTH - 1:
                data = data + self._extract(
                    AMQPHeartbeatFrame.LENGTH - 1 - len(data)).tobytes()

            if data != AMQPHeartbeatFrame.DATA[1:]:
                # Invalid heartbeat frame!
                raise ValueError('Invalid AMQP heartbeat')

            self.on_frame(AMQPHeartbeatFrame())
            self.frame_type = None

            return True

        # state rule 3
        elif (self.frame_type != FRAME_HEARTBEAT) and (
                self.frame_type is not None) and (
                self.frame_size is None) and (
                self.total_data_len > 6):
            hdr = b''
            while len(hdr) < 6:
                hdr = hdr + self._extract(6 - len(hdr)).tobytes()

            self.frame_channel, self.frame_size = struct.unpack('!HI', hdr)

            return True

        # state rule 4
        elif (self.frame_size is not None) and (
                self.total_data_len >= (self.frame_size + 1)):

            if len(self.chunks[0]) >= self.frame_size:
                # We can subslice it - it's very fast
                payload = self._extract(self.frame_size)
            else:
                # Construct a separate buffer :(
                payload = io.BytesIO()
                while payload.tell() < self.frame_size:
                    payload.write(
                        self._extract(self.frame_size - payload.tell()))

                assert payload.tell() <= self.frame_size

                payload = memoryview(payload.getvalue())

            if self._extract_single_byte() != FRAME_END:
                raise ValueError('Invalid frame end')

            try:
                frame = FRAME_TYPES[self.frame_type].unserialize(
                    self.frame_channel, payload)
            except ValueError:
                raise

            self.on_frame(frame)
            self.frame_type = None
            self.frame_size = None

            return True
        return False



# coding=UTF-8
from __future__ import print_function, absolute_import, division

import logging
import time
import typing as tp
import warnings
from concurrent.futures import Future

import six

from coolamqp.attaches import Publisher, AttacheGroup, Consumer, Declarer
from coolamqp.attaches.utils import close_future
from coolamqp.clustering.events import ConnectionLost, MessageReceived, \
    NothingMuch, Event
from coolamqp.clustering.single import SingleNodeReconnector
from coolamqp.exceptions import ConnectionDead
from coolamqp.objects import Exchange, Message, Queue, QueueBind
from coolamqp.uplink import ListenerThread
from coolamqp.utils import monotonic

logger = logging.getLogger(__name__)

nothing_much = NothingMuch()


# If any spans are spawn here, it's Cluster's job to finish them, except for publish()
class Cluster(object):
    """
    Frontend for your AMQP needs.

    This has ListenerThread.

    Call .start() to connect to AMQP.

    It is not safe to fork() after .start() is called, but it's OK before.

    :param nodes: list of nodes, or a single node. For now, only one is supported.
    :param on_fail: callable/0 to call when connection fails in an
        unclean way. This is a one-shot
    :param extra_properties: refer to documentation in [/coolamqp/connection/connection.py]
        Connection.__init__
    :param log_frames: an object that supports logging each and every frame CoolAMQP sends and
        receives from the broker
    :type log_frames: tp.Optional[:class:`coolamqp.tracing.BaseFrameTracer`]
    :param name: name to appear in log items and prctl() for the listener thread
    :param on_blocked: callable to call when ConnectionBlocked/ConnectionUnblocked is received. It will be
        called with a value of True if connection becomes blocked, and False upon an unblock
    :param tracer: tracer, if opentracing is installed
    """

    # Events you can be informed about
    ST_LINK_LOST = 0  # Link has been lost
    ST_LINK_REGAINED = 1  # Link has been regained

    def __init__(self, nodes,  # type: tp.Union[NodeDefinition, tp.List[NodeDefinition]]
                 on_fail=None,  # type: tp.Optional[tp.Callable[[], None]]
                 extra_properties=None,
                 # type: tp.Optional[tp.List[tp.Tuple[bytes, tp.Tuple[bytes, str]]]]
                 log_frames=None,
                 name=None,  # type: tp.Optional[str]
                 on_blocked=None,  # type: tp.Callable[[bool], None],
                 tracer=None  # type: opentracing.Traccer
                 ):
        from coolamqp.objects import NodeDefinition
        if isinstance(nodes, NodeDefinition):
            nodes = [nodes]

        if len(nodes) > 1:
            raise NotImplementedError(u'Multiple nodes not supported yet')

        if tracer is not None:
            try:
                import opentracing
            except ImportError:
                raise RuntimeError('tracer given, but opentracing is not installed!')

        self.started = False            # type: bool
        self.tracer = tracer
        self.name = name or 'CoolAMQP'  # type: str
        self.node, = nodes              # type: NodeDefinition
        self.extra_properties = extra_properties
        self.log_frames = log_frames
        self.on_blocked = on_blocked    # type: tp.Optional[tp.Callable[[bool], None]]
        self.connected = False          # type: bool
        self.listener = None            # type: BaseListener
        self.attache_group = None       # type: AttacheGroup
        self.events = None              # type: six.moves.queue.Queue
        self.snr = None                 # type: SingleNodeReconnector
        self.pub_tr = None              # type: Publisher
        self.pub_na = None              # type: Publisher
        self.decl = None                # type: Declarer

        if on_fail is not None:
            def decorated():
                if not self.listener.terminating and self.connected:
                    on_fail()

            self.on_fail = decorated
        else:
            self.on_fail = None

    def bind(self, queue, exchange, routing_key, persistent=False, span=None,
             dont_trace=False):
        """
        Bind a queue to an exchange
        """
        if span is not None and not dont_trace:
            child_span = self._make_span('bind', span)
        else:
            child_span = None
        fut = self.decl.declare(QueueBind(queue, exchange, routing_key),
                                persistent=persistent,
                                span=child_span)
        return close_future(fut, child_span)

    def declare(self, obj,  # type: tp.Union[Queue, Exchange]
                persistent=False,  # type: bool
                span=None,  # type: tp.Optional[opentracing.Span]
                dont_trace=False    # type: bool
                ):  # type: (...) -> concurrent.futures.Future
        """
        Declare a Queue/Exchange

        :param obj: Queue/Exchange object
        :param persistent: should it be redefined upon reconnect?
        :param span: optional parent span, if opentracing is installed
        :param dont_trace: if True, a span won't be output
        :return: Future
        """
        if span is not None and not dont_trace:
            child_span = self._make_span('declare', span)
        else:
            child_span = None
        fut = self.decl.declare(obj, persistent=persistent, span=child_span)
        return close_future(fut, child_span)

    def drain(self, timeout, span=None, dont_trace=False):  # type: (float) -> Event
        """
        Return an Event.

        :param timeout: time to wait for an event. 0 means return immediately. None means block forever
        :param span: optional parent span, if opentracing is installed
        :param dont_trace: if True, this span won't be traced
        :return: an Event instance. NothingMuch is returned when there's nothing within a given timoeout
        """

        def fetch():
            try:
                if timeout == 0:
                    return self.events.get_nowait()
                else:
                    return self.events.get(True, timeout)
            except six.moves.queue.Empty:
                return nothing_much

        if span is not None and not dont_trace:
            from opentracing import tags
            parent_span = self.tracer.start_active_span('AMQP call',
                                                        child_of=span,
                                                        tags={
                                                            tags.SPAN_KIND: tags.SPAN_KIND_RPC_CLIENT,
                                                            tags.DATABASE_TYPE: 'amqp',
                                                            tags.DATABASE_STATEMENT: 'drain'
                                                        })

            with parent_span:
                return fetch()
        else:
            return fetch()

    def consume(self, queue, on_message=None, span=None,
                dont_trace=False,   # type: bool
                *args, **kwargs):
        # type: (Queue, tp.Callable[[MessageReceived], None]) -> tp.Tuple[Consumer, Future]
        """
        Start consuming from a queue.

        args and kwargs will be passed to Consumer constructor (coolamqp.attaches.consumer.Consumer).
        Don't use future_to_notify - it's done here!

        Take care not to lose the Consumer object - it's the only way to cancel a consumer!

        :param queue: Queue object, being consumed from right now.
            Note that name of anonymous queue might change at any time!
        :param on_message: callable that will process incoming messages
                           if you leave it at None, messages will be .put into self.events
        :param span: optional span, if opentracing is installed
        :param dont_trace: if True, this won't output a span
        :return: a tuple (Consumer instance, and a Future), that tells, when consumer is ready
        """
        if span is not None and not dont_trace:
            child_span = self._make_span('consume', span)
        else:
            child_span = None
        fut = Future()
        fut.set_running_or_notify_cancel()  # it's running right now
        on_message = on_message or (
            lambda msg: self.events.put_nowait(MessageReceived(msg)))
        con = Consumer(queue, on_message, future_to_notify=fut, span=span, *args,
                       **kwargs)
        self.attache_group.add(con)
        return con, close_future(fut, child_span)

    def delete_queue(self, queue):  # type: (coolamqp.objects.Queue) -> Future
        """
        Delete a queue.

        :param queue: Queue instance that represents what to delete
        :return: a Future (will succeed with None or fail with AMQPError)
        """
        return self.decl.delete_queue(queue)

    def _make_span(self, call, span):
        try:
            from opentracing import tags
        except ImportError:
            pass
        else:
            return self.tracer.start_span('AMQP call',
                                          child_of=span,
                                          tags={
                                              tags.SPAN_KIND: tags.SPAN_KIND_RPC_CLIENT,
                                              tags.DATABASE_TYPE: 'amqp',
                                              tags.DATABASE_STATEMENT: call
                                          })

    def publish(self, message,  # type: Message
                exchange=None,  # type: tp.Union[Exchange, str, bytes]
                routing_key=u'',  # type: tp.Union[str, bytes]
                tx=None,  # type: tp.Optional[bool]
                confirm=None,  # type: tp.Optional[bool]
                span=None,  # type: tp.Optional[opentracing.Span]
                dont_trace=False    # type: bool
                ):  # type: (...) -> tp.Optional[Future]
        """
        Publish a message.

        :param message: Message to publish
        :param exchange: exchange to use. Default is the "direct" empty-name exchange.
        :param routing_key: routing key to use
        :param confirm: Whether to publish it using confirms/transactions.
                        If you choose so, you will receive a Future that can be used
                        to check it broker took responsibility for this message.
                        Note that if tx if False, and message cannot be delivered to broker at once,
                        it will be discarded
        :param tx: deprecated, alias for confirm
        :param span: optionally, current span, if opentracing is installed
        :param dont_trace: if set to True, a span won't be generated
        :return: Future to be finished on completion or None, is confirm/tx was not chosen
        """
        if self.tracer is not None and not dont_trace:
            span = self._make_span('publish', span)

        if isinstance(exchange, Exchange):
            exchange = exchange.name.encode('utf8')
        elif exchange is None:
            exchange = b''
        elif isinstance(exchange, six.text_type):
            exchange = exchange.encode('utf8')

        if isinstance(routing_key, six.text_type):
            routing_key = routing_key.encode('utf8')

        if tx is not None:  # confirm is a drop-in replacement. tx is unfortunately named
            warnings.warn(u'Use confirm kwarg instead', DeprecationWarning)

            if confirm is not None:
                raise RuntimeError(
                    u'Using both tx= and confirm= at once does not make sense')
        elif confirm is not None:
            tx = confirm
        else:
            tx = False

        try:
            if tx:
                clb = self.pub_tr
            else:
                clb = self.pub_na
            return clb.publish(message, exchange, routing_key, span)
        except Publisher.UnusablePublisher:
            raise NotImplementedError(
                u'Sorry, this functionality is not yet implemented!')

    def start(self, wait=True, timeout=10.0):  # type: (bool, float, bool) -> None
        """
        Connect to broker. Initialize Cluster.

        Only after this call is Cluster usable.
        It is not safe to fork after this.

        :param wait: block until connection is ready
        :param timeout: timeout to wait until the connection is ready. If it is not, a
                        ConnectionDead error will be raised
        :raise RuntimeError: called more than once
        :raise ConnectionDead: failed to connect within timeout
        """
        if self.started:
            raise RuntimeError(u'[%s] This was already called!' % (self.name,))
        self.started = True

        self.listener = ListenerThread(name=self.name)

        self.attache_group = AttacheGroup()

        self.events = six.moves.queue.Queue()  # for coolamqp.clustering.events.*

        self.snr = SingleNodeReconnector(self.node, self.attache_group,
                                         self.listener, self.extra_properties,
                                         self.log_frames, self.name)
        self.snr.on_fail.add(lambda: self.events.put_nowait(ConnectionLost()))
        if self.on_fail is not None:
            self.snr.on_fail.add(self.on_fail)

        if self.on_blocked is not None:
            self.snr.on_blocked.add(self.on_blocked)

        # Spawn a transactional publisher and a noack publisher
        self.pub_tr = Publisher(Publisher.MODE_CNPUB, self)
        self.pub_na = Publisher(Publisher.MODE_NOACK, self)
        self.decl = Declarer(self)

        self.attache_group.add(self.pub_tr)
        self.attache_group.add(self.pub_na)
        self.attache_group.add(self.decl)

        self.listener.init()
        self.listener.start()
        self.snr.connect(timeout=timeout)

        if wait:
            # this is only going to take a short amount of time, so we're fine with polling
            start_at = monotonic()
            while not self.connected and monotonic() - start_at < timeout:
                time.sleep(0.1)
            if not self.connected:
                raise ConnectionDead(
                    '[%s] Could not connect within %s seconds' % (self.name, timeout,))

    def shutdown(self, wait=True):  # type: (bool) -> None
        """
        Terminate all connections, release resources - finish the job.

        :param wait: block until this is done
        :raise RuntimeError: if called without start() being called first
        """
        self.connected = False
        if not self.started:
            raise RuntimeError(u'shutdown without start')

        logger.info('[%s] Commencing shutdown', self.name)

        self.listener.terminate()
        if wait:
            self.listener.join()



# coding=UTF-8
from __future__ import print_function, absolute_import, division

import logging
import typing as tp

from coolamqp.framing.definitions import ConnectionUnblocked, ConnectionBlocked
from coolamqp.objects import Callable
from coolamqp.uplink import Connection
from coolamqp.uplink.connection import MethodWatch

logger = logging.getLogger(__name__)


class SingleNodeReconnector(object):
    """
    Connection to one node. It will do it's best to remain alive.
    """

    def __init__(self, node_def,  # type: coolamqp.objects.NodeDefinition
                 attache_group,  # type: coolamqp.attaches.AttacheGroup
                 listener_thread,  # type: coolamqp.uplink.ListenerThread
                 extra_properties=None,  # type: tp.Dict[bytes, tp.Tuple[tp.Any, str]]
                 log_frames=None,  # type: tp.Callable[]
                 name=None):
        self.listener_thread = listener_thread
        self.node_def = node_def
        self.attache_group = attache_group
        self.connection = None
        self.extra_properties = extra_properties
        self.log_frames = log_frames
        self.name = name or 'CoolAMQP'

        self.terminating = False
        self.timeout = None

        self.on_fail = Callable()  #: public
        self.on_blocked = Callable()  #: public
        self.on_fail.add(self._on_fail)

    def is_connected(self):  # type: () -> bool
        return self.connection is not None

    def connect(self, timeout=None):  # type: (tp.Optional[float]) -> None
        assert self.connection is None

        timeout = timeout or self.timeout
        self.timeout = timeout

        # Initiate connecting - this order is very important!
        self.connection = Connection(self.node_def, self.listener_thread,
                                     extra_properties=self.extra_properties,
                                     log_frames=self.log_frames,
                                     name=self.name)
        self.attache_group.attach(self.connection)
        self.connection.start(timeout)
        self.connection.finalize.add(self.on_fail)

        # Register the on-blocking watches
        mw = MethodWatch(0, (ConnectionBlocked,), lambda: self.on_blocked(True))
        mw.oneshot = False
        self.connection.watch(mw)

        mw = MethodWatch(0, (ConnectionUnblocked,), lambda: self.on_blocked(False))
        mw.oneshot = False
        self.connection.watch(mw)

    def _on_fail(self):
        if self.terminating:
            return

        self.connection = None
        self.listener_thread.call_next_io_event(self.connect)

    def shutdown(self):
        """Close this connection"""
        self.terminating = True

        if self.connection is not None:
            self.connection.send(None)
            self.connection = None



# coding=UTF-8
from __future__ import absolute_import, division, print_function

import io


class SendingFramer(object):
    """
    Assembles AMQP framing from received data and orchestrates their upload via a socket.

    Just call with .put(data) and get framing by iterator .framing().

    Not thread safe.

    State machine
        (frame_type is None)  and has_bytes(1)        ->          (frame_type <- bytes(1))

        (frame_type is HEARTBEAT) and has_bytes(3)  ->          (output_frame, frame_type <- None)
        (frame_type is not HEARTBEAT and not None) and has_bytes(6)  ->      (frame_channel <- bytes(2),
                                                                 frame_size <- bytes(4))

        (frame_size is not None) and has_bytes(frame_size+1)    ->  (output_frame,
                                                                            frame_type <- None
                                                                            frame_size < None)
    """

    def __init__(self, on_send=lambda data: None):
        """
        :param on_send: a callable(data, priority=False) that can be called with some data to send
            data will always be entire AMQP frames!
        """
        self.on_send = on_send

    def send(self, frames, priority=False):
        """
        Schedule to send some frames.
        :param frames: list of AMQPFrame instances
        :param priority: preempty existing frames
        """
        length = sum(frame.get_size() for frame in frames)
        buf = io.BytesIO(bytearray(length))

        for frame in frames:
            frame.write_to(buf)

        q = buf.getvalue()
        self.on_send(q, priority)



# coding=UTF-8
from __future__ import absolute_import, division, print_function

from coolamqp.framing.field_table import deframe_table

"""Generate serializers/unserializers/length getters for given property_flags"""
import six
import struct
import logging
from coolamqp.framing.compilation.textcode_fields import get_counter, \
    get_from_buffer, get_serializer

logger = logging.getLogger(__name__)

INIT_I = u'\n    def __init__(self, %s):\n'
SLOTS_I = u'\n    __slots__ = (%s)\n'
FROM_BUFFER_1 = u'    def from_buffer(cls, buf, start_offset):\n        ' \
                u'offset = start_offset + %s\n'
ASSIGN_A = u'        self.%s = %s\n'
STARTER = u'''from coolamqp.framing.base import AMQPContentPropertyList

class ParticularContentTypeList(AMQPContentPropertyList):
    """
    For fields:
'''
ZPF_S = u'''
    # A value for property flags that is used, assuming all bit fields are FALSE (0)
    ZERO_PROPERTY_FLAGS = %s
'''
NB = u"raise NotImplementedError('I don't support bits in properties')"
INTER_X = u'    * %s::%s'
BUF_WRITE_A = u'\n    def write_to(self, buf):\n        buf.write('
RESERVED = u' (reserved)'
UNICO = u"u'%s'"
SPACER = u'''
    """
'''
GET_SIZE_HEADER = u'\n    def get_size(self):\n'


def _compile_particular_content_property_list_class(zpf, fields):
    """
    Compile a particular content property list.

    Particularity stems from
    :param zpf: zero property list, as bytearray
    :param fields: list of all possible fields in this content property
    """
    from coolamqp.framing.compilation.utilities import format_field_name

    structers = {}

    if any(field.basic_type == 'bit' for field in fields):
        return NB

    # Convert ZPF to a list of [if_exists::bool]
    even = True
    zpf_bits = []
    for q in bytearray(zpf):
        p = bin(q)[2:]
        p = (u'0' * (8 - len(p))) + p

        if not even:
            p = p[:7]

        for x in p:
            zpf_bits.append(bool(int(x)))

    zpf_length = len(zpf)

    # 1 here does not mean that field is present. All bit fields are present,
    # but 0 in a ZPF. Fix this.
    zpf_bits = [zpf_bit or field.type == 'bit' for zpf_bit, field in
                zip(zpf_bits, fields)]

    mod = [STARTER]

    for field in fields:
        mod.append(
            INTER_X % (format_field_name(field.name), field.type))
        if field.reserved:
            mod.append(RESERVED)
        mod.append(u'\n')

    x = repr(six.binary_type(zpf))
    if not x.startswith('b'):
        x = 'b' + x

    present_fields = [field for field, present in zip(fields, zpf_bits) if
                      present]

    mod.append(SPACER)

    if len(present_fields) == 0:
        slots = u''
    else:
        slots = (u', '.join(
            (UNICO % format_field_name(field.name) for field in
             present_fields))) + u', '

    mod.append(SLOTS_I % slots)

    mod.append(ZPF_S % (x,))

    FFN = u', '.join(format_field_name(field.name) for field in present_fields)

    if len(present_fields) > 0:
        mod.append(INIT_I % (FFN,))

    for field in present_fields:
        mod.append(ASSIGN_A.replace(u'%s', format_field_name(
            field.name)))

    # Let's do write_to
    mod.append(BUF_WRITE_A)
    repred_zpf = repr(zpf)
    if not repred_zpf.startswith(u'b'):
        repred_zpf = u'b' + repred_zpf
    mod.append(repred_zpf)
    mod.append(u')\n')

    line, new_structers = get_serializer(present_fields, prefix=u'self.', indent_level=2)
    structers.update(new_structers)
    mod.append(line)

    # from_buffer
    # note that non-bit values
    mod.append(u'    @classmethod\n')
    mod.append(
        FROM_BUFFER_1 % (
            zpf_length,))
    line, new_structers = get_from_buffer(
        present_fields
        , prefix='', indent_level=2)
    structers.update(new_structers)
    mod.append(line)
    mod.append(u'        return cls(%s)\n' % (FFN,))

    # get_size
    mod.append(GET_SIZE_HEADER)
    mod.append(get_counter(present_fields, prefix=u'self.', indent_level=2)[
               :-1])  # skip eol
    mod.append(u' + %s\n' % (zpf_length,))  # account for pf length

    return u''.join(mod), structers


STRUCTERS_FOR_NOW = {}  # type: tp.Dict[str, struct.Struct]


def compile_particular_content_property_list_class(zpf, fields):
    from coolamqp.framing.base import AMQPContentPropertyList
    global STRUCTERS_FOR_NOW

    q, structers = _compile_particular_content_property_list_class(zpf, fields)
    locals_ = {
        'AMQPContentPropertyList': AMQPContentPropertyList,
        'deframe_table': deframe_table,
    }
    for structer in structers:
        if structer not in STRUCTERS_FOR_NOW:
            STRUCTERS_FOR_NOW[structer] = struct.Struct('!%s' % (structer,))

        locals_['STRUCT_%s' % (structer,)] = STRUCTERS_FOR_NOW[structer]

    loc = dict(globals(), **locals_)
    exec(q, loc)
    return loc['ParticularContentTypeList']



# coding=utf-8
"""
Module used to publish messages.

Expect wild NameErrors if you build this without RabbitMQ extensions (enabled by default),
and try to use MODE_CNPUB.

If you use a broker that doesn't support these, just don't use MODE_CNPUB. CoolAMQP is smart enough
to check with the broker beforehand.
"""
from __future__ import absolute_import, division, print_function

import collections
import logging
import warnings

import six

from coolamqp.framing.definitions import ChannelOpenOk, BasicPublish, Basic, \
    BasicAck
from coolamqp.framing.frames import AMQPMethodFrame, AMQPBodyFrame, \
    AMQPHeaderFrame

try:
    # these extensions will be available
    from coolamqp.framing.definitions import ConfirmSelect, ConfirmSelectOk, \
        BasicNack, ChannelFlow, ChannelFlowOk, ConnectionBlocked, ConnectionUnblocked
except ImportError:
    pass

from coolamqp.attaches.channeler import Channeler, ST_ONLINE, ST_OFFLINE
from coolamqp.uplink import PUBLISHER_CONFIRMS, MethodWatch, FailWatch
from coolamqp.attaches.utils import AtomicTagger, FutureConfirmableRejectable, \
    Synchronized

from concurrent.futures import Future
from coolamqp.objects import Exchange

logger = logging.getLogger(__name__)

# for holding messages when MODE_CNPUB and link is down
CnpubMessageSendOrder = collections.namedtuple('CnpubMessageSendOrder',
                                               ('message', 'exchange_name',
                                                'routing_key', 'future',
                                                'parent_span', 'span_enqueued'))


# todo what if publisher in MODE_CNPUB fails mid message? they dont seem to be recovered


class Publisher(Channeler, Synchronized):
    """
    An object that is capable of sucking into a Connection and sending messages.
    Depending on it's characteristic, it may process messages in:

        - non-ack mode (default) - messages will be dropped on the floor if there is no active uplink
        - Consumer Publish mode - requires broker support, each message will be ACK/NACKed by the broker
                                  messages will survive broker reconnections.

                                  If you support this, it is your job to ensure that broker supports
                                  publisher_confirms. If it doesn't, this publisher will enter ST_OFFLINE
                                  and emit a warning.

        Other modes may be added in the future.

    Since this may be called by other threads than ListenerThread, this has locking.

    _pub and on_fail are synchronized so that _pub doesn't see a partially destroyed class.

    :param mode: Publishing mode to use. One of:
         MODE_NOACK - use non-ack mode
         MODE_CNPUB - use consumer publishing mode. A switch to MODE_TXPUB will be made
                      if broker does not support these.
    :raise ValueError: mode invalid
    """
    MODE_NOACK = 0  # no-ack publishing
    MODE_CNPUB = 1  # RabbitMQ publisher confirms extension

    # todo add fallback using plain AMQP transactions - this will remove UnusablePublisher and stuff

    class UnusablePublisher(Exception):
        """This publisher will never work (eg. MODE_CNPUB on a broker not supporting publisher confirms)"""

    def __init__(self, mode, cluster=None):
        Channeler.__init__(self)
        Synchronized.__init__(self)

        if mode not in (Publisher.MODE_NOACK, Publisher.MODE_CNPUB):
            raise ValueError(u'Invalid publisher mode')

        self.mode = mode

        self.messages = collections.deque()  # Messages to publish. From newest to last.
        # tuple of (Message object, exchange name::str, routing_key::str,
        #           Future to confirm or None, flags as tuple|empty tuple

        self.tagger = None  # None, or AtomicTagger instance id MODE_CNPUB
        self.set_connected = False
        self.cluster = cluster
        self.critically_failed = False
        self.content_flow = True
        self.blocked = False
        self.frames_to_send = []

    @Synchronized.synchronized
    def attach(self, connection):
        Channeler.attach(self, connection)
        connection.watch(FailWatch(self.on_fail))

    def on_connection_blocked(self, payload):
        if isinstance(payload, ConnectionBlocked):
            self.blocked = True
        elif isinstance(payload, ConnectionUnblocked):
            self.blocked = False

            if self.content_flow:
                self.connection.send(self.frames_to_send)
                self.frames_to_send = []

    def on_flow_control(self, payload):
        """Called on ChannelFlow"""
        assert isinstance(ChannelFlow, payload)

        self.content_flow = payload.active
        self.connection.send([AMQPMethodFrame(self.channel_id,
                                              ChannelFlowOk(payload.active))])

        if payload.active and not self.blocked:
            self.connection.send(self.frames_to_send)
            self.frames_to_send = []

    @Synchronized.synchronized
    def on_fail(self):
        self.state = ST_OFFLINE

    def _pub(self, message, exchange_name, routing_key, parent_span=None, span_enqueued=None,
             dont_close_span=False):
        """
        Just send the message. Sends BasicDeliver + header + body.

        BECAUSE OF publish THIS CAN GET CALLED BY FOREIGN THREAD.

        :param message: Message instance
        :param exchange_name: exchange to use
        :param routing_key: routing key to use
        :type exchange_name: bytes
        :param routing_key: bytes
        """
        span = None
        if parent_span is not None:
            import opentracing
            span_enqueued.finish()
            span = self.cluster.tracer.start_span('Sending',
                                                  child_of=parent_span,
                                                  references=opentracing.follows_from(span_enqueued))
        # Break down large bodies
        bodies = []

        body = memoryview(message.body)
        max_body_size = self.connection.frame_max - AMQPBodyFrame.FRAME_SIZE_WITHOUT_PAYLOAD - 16
        while len(body) > 0:
            bodies.append(body[:max_body_size])
            body = body[max_body_size:]

        frames_to_send = [AMQPMethodFrame(self.channel_id,
                                          BasicPublish(exchange_name, routing_key, False, False)),
                          AMQPHeaderFrame(self.channel_id, Basic.INDEX, 0, len(message.body),
                                          message.properties)]

        if len(bodies) == 1:
            frames_to_send.append(AMQPBodyFrame(self.channel_id, bodies[0]))

        if self.content_flow and not self.blocked:
            self.connection.send(frames_to_send)

            if len(bodies) > 1:
                while self.content_flow and not self.blocked and len(bodies) > 0:
                    self.connection.send([AMQPBodyFrame(self.channel_id, bodies[0])])
                    del bodies[0]

                if not self.content_flow and not self.blocked and len(bodies) > 0:
                    for body in bodies:
                        self.frames_to_send.append(AMQPBodyFrame(self.channel_id, body))
        else:
            self.frames_to_send.extend(frames_to_send)
            if len(bodies) > 1:
                for body in bodies:
                    self.frames_to_send.append(AMQPBodyFrame(self.channel_id, body))

        if span is not None:
            span.finish()

        if parent_span is not None and not dont_close_span:
            parent_span.finish()

    def _mode_cnpub_process_deliveries(self):
        """
        Dispatch all frames that are waiting to be sent

        To be used when  mode is MODE_CNPUB and we just got ST_ONLINE
        """
        assert self.state == ST_ONLINE
        assert self.mode == Publisher.MODE_CNPUB
        assert self.tagger is not None

        while len(self.messages) > 0:
            try:
                msg, xchg, rk, fut, parent_span, span_enqueued = self.messages.popleft()
            except IndexError:
                # todo see docs/casefile-0001
                break

            if not fut.set_running_or_notify_cancel():
                if span_enqueued is not None:
                    from opentracing import logs
                    span_enqueued.log_kv({logs.EVENT: 'Cancelled'})
                    span_enqueued.finish()
                    parent_span.finish()
                continue  # cancelled

            self.tagger.deposit(self.tagger.get_key(),
                                FutureConfirmableRejectable(fut),
                                parent_span)
            assert isinstance(xchg, (six.binary_type, six.text_type))
            self._pub(msg, xchg, rk, parent_span, span_enqueued, dont_close_span=True)

    def _on_cnpub_delivery(self, payload):  # type: (AMQPMethodPayload) -> None
        """
        This gets called on BasicAck and BasicNack, if mode is MODE_CNPUB
        """
        assert self.mode == Publisher.MODE_CNPUB

        if isinstance(payload, BasicAck):
            self.tagger.ack(payload.delivery_tag, payload.multiple)
        elif isinstance(payload, BasicNack):
            self.tagger.nack(payload.delivery_tag, payload.multiple)

    @Synchronized.synchronized
    def publish(self, message, exchange=b'', routing_key=b'', span=None):
        """
        Schedule to have a message published.

        If mode is MODE_CNPUB:
            this function will return a Future. Future can end either with success (result will be None),
            or exception (a plain Exception instance). Exception will happen when broker NACKs the message:
            that, according to RabbitMQ, means an internal error in Erlang process.

            Returned Future can be cancelled - this will prevent from sending the message, if it hasn't commenced yet.

        If mode is MODE_NOACK:
            this function returns None. Messages are dropped on the floor if there's no connection.

        :param message: Message object to send
        :param exchange: exchange name to use. Default direct exchange by default. Can also be an Exchange object.
        :type exchange: bytes, str or Exchange instance
        :param routing_key: routing key to use
        :param span: optional span, if opentracing is installed
        :return: a Future instance, or None
        :raise Publisher.UnusablePublisher: this publisher will never work (eg. MODE_CNPUB on Non-RabbitMQ)
        """
        if span is not None:
            span_enqueued = self.cluster.tracer.start_span('Enqueued', child_of=span)
        else:
            span_enqueued = None

        if isinstance(exchange, Exchange):
            exchange = exchange.name.encode('utf8')
        elif isinstance(exchange, six.text_type):
            exchange = exchange.encode('utf8')

        assert isinstance(exchange, six.binary_type)

        # Formulate the request
        if self.mode == Publisher.MODE_NOACK:
            # If we are not connected right now, drop the message on the floor and log it with DEBUG
            if self.state != ST_ONLINE:
                logger.debug(
                    u'Publish request, but not connected - dropping the message')
            else:
                self._pub(message, exchange, routing_key, span, span_enqueued)

        elif self.mode == Publisher.MODE_CNPUB:
            fut = Future()

            # todo can optimize this not to create an object if ST_ONLINE already
            cnpo = CnpubMessageSendOrder(message, exchange, routing_key, fut, span, span_enqueued)
            self.messages.append(cnpo)

            if self.state == ST_ONLINE:
                self._mode_cnpub_process_deliveries()

            return fut
        else:
            raise Exception(u'Invalid mode')

    def on_operational(self, operational):      # type: (bool) -> None
        state = {True: u'up', False: u'down'}[operational]
        mode = \
            {Publisher.MODE_NOACK: u'noack', Publisher.MODE_CNPUB: u'cnpub'}[
                self.mode]

        logger.info('Publisher %s is %s', mode, state)

    def on_setup(self, payload):

        # Assert that mode is OK
        if self.mode == Publisher.MODE_CNPUB:
            if PUBLISHER_CONFIRMS not in self.connection.extensions:
                warnings.warn(
                    u'Broker does not support publisher_confirms, refusing to start publisher',
                    RuntimeWarning)
                self.state = ST_OFFLINE
                self.critically_failed = True
                return

        logger.debug('Publisher on_setup, payload=%s', payload)

        if isinstance(payload, ChannelOpenOk):
            # Ok, if this has a mode different from MODE_NOACK, we need to additionally set up
            # the functionality.
            mw = self.watch_for_method(ChannelFlow, self.on_flow_control)
            mw.oneshot = False

            mw = self.connection.watch_for_method(0, (ConnectionBlocked, ConnectionUnblocked),
                                                  self.on_connection_blocked)
            mw.oneshot = False

            if self.mode == Publisher.MODE_CNPUB:
                self.method_and_watch(ConfirmSelect(False), ConfirmSelectOk,
                                      self.on_setup)
            elif self.mode == Publisher.MODE_NOACK:
                # A-OK! Boot it.
                self.state = ST_ONLINE
                self.on_operational(True)

        elif (self.mode == Publisher.MODE_CNPUB) and isinstance(payload, ConfirmSelectOk):
            # Because only in this case it makes sense to check for MODE_CNPUB
            # A-OK! Boot it.
            self.tagger = AtomicTagger()
            self.state = ST_ONLINE
            self.on_operational(True)

            # inform the cluster that we've been connected
            if not self.set_connected:
                self.cluster.connected = True
                self.set_connected = True

            # now we need to listen for BasicAck and BasicNack

            mw = MethodWatch(self.channel_id, (BasicAck, BasicNack),
                             self._on_cnpub_delivery)
            mw.oneshot = False
            self.connection.watch(mw)
            self._mode_cnpub_process_deliveries()




try:
    IMPORT_ERRORS = (ModuleNotFoundError, ImportError)
except NameError:
    IMPORT_ERRORS = (ImportError, )

try:
    from time import monotonic
except IMPORT_ERRORS:
    from monotonic import monotonic

try:
    from prctl import set_name as prctl_set_name
except IMPORT_ERRORS:
    def prctl_set_name(name):
        pass


__all__ = ['monotonic', 'prctl_set_name']



# coding=UTF-8
"""
This is the layer that you talk to. It abstracts away one (in future - more) connections
to broker with an uniform interface.
"""
from __future__ import print_function, absolute_import, division

from coolamqp.clustering.cluster import Cluster
from coolamqp.clustering.events import MessageReceived, NothingMuch, \
    ConnectionLost

__all__ = ['Cluster', 'MessageReceived', 'NothingMuch', 'ConnectionLost']



# coding=UTF-8
from __future__ import absolute_import, division, print_function

import logging
import select
import socket
import threading

import six

from coolamqp.uplink.listener.socket import SocketFailed, BaseSocket
from coolamqp.uplink.listener.base_listener import BaseListener


logger = logging.getLogger(__name__)

try:
    RO = select.EPOLLIN | select.EPOLLHUP | select.EPOLLERR
    RW = RO | select.EPOLLOUT
except AttributeError:
    # epoll listener will be unusable anyway
    RO = 0
    RW = 1


class EpollSocket(BaseSocket):

    def send(self, data, priority=False):
        """
        This can actually get called not by ListenerThread.
        """
        BaseSocket.send(self, data, priority=priority)
        try:
            self.listener.epoll.modify(self, RW)
        except socket.error:
            # silence. If there are errors, it's gonna get nuked soon.
            pass


class EpollListener(BaseListener):
    """
    A listener using epoll.
    """

    def __init__(self):
        self.epoll = select.epoll()
        self.socket_activation_lock = threading.Lock()
        self.sockets_to_activate = []
        super(EpollListener, self).__init__()

    def wait(self, timeout=1):
        with self.socket_activation_lock:
            for socket_to_activate in self.sockets_to_activate:
                logger.debug('Activating fd %s', (socket_to_activate.fileno(),))
                self.epoll.register(socket_to_activate.fileno(), RW)
            self.sockets_to_activate = []

        events = self.epoll.poll(timeout=timeout)

        self.do_timer_events()

        for fd, event in events:
            sock = self.fd_to_sock[fd]

            # Errors
            try:
                if event & (select.EPOLLERR | select.EPOLLHUP):
                    logger.debug('Socket %s has failed', fd)
                    raise SocketFailed()

                if event & select.EPOLLIN:
                    sock.on_read()

                if event & select.EPOLLOUT:
                    sock.on_write()
                    # I'm done with sending for now
                    if len(sock.data_to_send) == 0 and len(
                            sock.priority_queue) == 0:
                        self.epoll.modify(sock.fileno(), RO)

            except SocketFailed as e:
                logger.debug('Socket %s has raised %s', fd, e)
                self.close_socket(sock)

        # Do any of the sockets want to send data Re-register them
        for sock in six.itervalues(self.fd_to_sock):
            if sock.wants_to_send_data():
                self.epoll.modify(sock.fileno(), RW)

    def close_socket(self, sock):  # type: (BaseSocket) -> None
        self.epoll.unregister(sock.fileno())
        super(EpollListener, self).close_socket(sock)

    def shutdown(self):
        """
        Forcibly close all sockets that this manages (calling their on_fail's),
        and close the object.

        This object is unusable after this call.
        """
        super(EpollListener, self).shutdown()
        self.epoll.close()

    def activate(self, sock):  # type: (BaseSocket) -> None
        super(EpollListener, self).activate(sock)
        with self.socket_activation_lock:
            self.sockets_to_activate.append(sock)

    def register(self, sock, on_read=lambda data: None,
                 on_fail=lambda: None):
        """
        Add a socket to be listened for by the loop.

        Please note that .activate() will be later called on this socket.

        :param sock: a socket instance (as returned by socket module)
        :param on_read: callable(data) to be called with received data
        :param on_fail: callable() to be called when socket fails

        :return: a BaseSocket instance to use instead of this socket
        """
        return EpollSocket(sock, on_read, on_fail=on_fail, listener=self)



# coding=UTF-8
from __future__ import absolute_import, division, print_function

import threading
import logging
import typing as tp
import os
from coolamqp.uplink.listener.select_listener import SelectListener
from coolamqp.objects import Callable
from coolamqp.uplink.listener.base_listener import BaseListener
from coolamqp.utils import prctl_set_name

logger = logging.getLogger(__name__)


def get_listener_class():   # type: () -> tp.Type[BaseListener]

    if 'COOLAMQP_FORCE_SELECT_LISTENER' in os.environ:
        return SelectListener

    try:
        import select
        select.EPOLLIN
        select.epoll
    except AttributeError:
        return SelectListener   # we're running on a platform that doesn't support epoll

    try:
        import gevent.socket
    except ImportError:
        from coolamqp.uplink.listener.epoll_listener import EpollListener
        return EpollListener
    import socket

    if socket.socket is gevent.socket.socket:
        return SelectListener     # gevent is active

    from coolamqp.uplink.listener.epoll_listener import EpollListener
    return EpollListener


class ListenerThread(threading.Thread):
    """
    A thread that does the listening.

    It automatically picks the best listener for given platform.
    """

    def __init__(self, name=None):  # type: (tp.Optional[str])
        super(ListenerThread, self).__init__(name=name or 'coolamqp/ListenerThread')
        self.daemon = True
        self.name = name or 'CoolAMQP'
        self.terminating = False
        self._call_next_io_event = Callable(oneshots=True)
        self.listener = None        # type: BaseListener

    def call_next_io_event(self, callable):
        """
        Call callable after current I/O event is fully processed

        sometimes many callables are called in response to single
        I/O (eg. teardown, startup). This guarantees a call after
        all these are done.
        :param callable: callable/0
        """
        pass
#        self._call_next_io_event.add(callable) - dummy that out, causes AssertionError to appear

    def terminate(self):
        self.terminating = True

    def init(self):
        """Called before start. It is not safe to fork after this"""
        listener_class = get_listener_class()
        logger.info('Using %s as a listener' % (listener_class, ))
        self.listener = listener_class()

    def activate(self, sock):
        self.listener.activate(sock)

    def run(self):
        prctl_set_name(self.name + '- listener thread')

        while not self.terminating:
            self.listener.wait()
            self._call_next_io_event()

        self.listener.shutdown()

    def register(self, sock,  # type: socket.socket
                 on_read=lambda data: None,  # type: tp.Callable[[bytes], None]
                 on_fail=lambda: None      # type: tp.Callable[[], None]
                 ):
        """
        Add a socket to be listened for by the loop.

        :param sock: a socket instance (as returned by socket module)
        :param on_read: callable(data) to be called with received data
        :param on_fail: callable() to be called when socket fails

        :return: a BaseSocket instance to use instead of this socket
        """
        return self.listener.register(sock, on_read, on_fail)



# coding=UTF-8
"""
This is an attache that attaches multiple attaches.

It evicts cancelled attaches.
"""
from __future__ import print_function, absolute_import, division

import logging

logger = logging.getLogger(__name__)

from coolamqp.attaches.channeler import Attache, ST_OFFLINE, ST_ONLINE
from coolamqp.attaches.consumer import Consumer
from coolamqp.attaches.publisher import Publisher


class AttacheGroup(Attache):
    """
    A bunch of attaches
    """

    def __init__(self):
        super(AttacheGroup, self).__init__()
        self.attaches = []

        # these two to be filled in during add()
        self.tx_publisher = None
        self.non_tx_publisher = None

    def add(self, attache):
        """
        Add an attache to this group.

        If this is attached, and connection is ST_ONLINE, .attach() will be called
        on this attache at once.

        :param attache: Attache instance
        """
        assert attache not in self.attaches
        self.attaches.append(attache)

        # If we have any connection, and it's not dead, attach
        if self.connection is not None and self.connection.state != ST_OFFLINE:
            attache.attach(self.connection)

        if isinstance(attache, Consumer):
            attache.attache_group = self

        if isinstance(attache, Publisher):
            if attache.mode == Publisher.MODE_CNPUB:
                self.tx_publisher = attache
            else:
                self.non_tx_publisher = attache

    def on_cancel_customer(self, customer):
        """
        Called by a customer, when it's cancelled.

        Consumer must have .attache_group set to this. This is done by .add()

        :param customer: a Customer instance
        """
        self.attaches.remove(customer)

    def attach(self, connection):
        """
        Attach to a connection

        :param connection: Connection instance of any state
        """
        # since this attache does not watch for failures, it can't use typical method.
        self.connection = connection

        for attache in self.attaches:
            if not attache.cancelled:
                attache.attach(connection)

    def is_online(self):  # type: () -> bool
        return self.tx_publisher.state == ST_ONLINE and self.non_tx_publisher.state == ST_ONLINE



# coding=UTF-8
"""

Core object here is Connection. This package:
    - establishes basic connectivity (up to the point where you can open channels yourself)
    - takes care of heartbeats

You can wait for a particular frame by setting watches on connections.
Watches will fire upon an event triggering them.

EVERYTHING HERE IS CALLED BY LISTENER THREAD UNLESS STATED OTHERWISE.

"""
from __future__ import absolute_import, division, print_function

from coolamqp.uplink.connection import Connection, HeaderOrBodyWatch, \
    MethodWatch, AnyWatch, FailWatch
from coolamqp.uplink.handshake import PUBLISHER_CONFIRMS, \
    CONSUMER_CANCEL_NOTIFY
from coolamqp.uplink.listener import ListenerThread



from abc import ABCMeta, abstractmethod
import heapq
import typing as tp
import six
from coolamqp.utils import monotonic


class BaseListener(object):
    __metaclass__ = ABCMeta

    def __init__(self):
        self.fd_to_sock = {}    # type: tp.Dict[int, BaseSocket]
        self.time_events = []  # type: tp.List[tp.Tuple[float, int, tp.Callable[[], None]]]

    def do_timer_events(self):
        # Timer events
        mono = monotonic()
        while len(self.time_events) > 0 and (self.time_events[0][0] < mono):
            ts, fd, callback = heapq.heappop(self.time_events)
            callback()

    def oneshot(self, sock, delta, callback):
        """
        A socket registers a time callback
        :param sock: BaseSocket instance
        :param delta: "this seconds after now"
        :param callback: callable/0
        """
        if sock.fileno() in self.fd_to_sock:
            heapq.heappush(self.time_events, (monotonic() + delta,
                                              sock.fileno(),
                                              callback
                                              ))

    def noshot(self, sock):     # type: (BaseSocket) -> None
        """
        Clear all one-shots for a socket
        :param sock: BaseSocket instance
        """
        fd = sock.fileno()
        self.time_events = [q for q in self.time_events if q[1] != fd]

    @abstractmethod
    def wait(self, timeout=1):
        """
        This will be executed in a loop.

        This must call .do_timer_events()
        """

    def close_socket(self, sock):   # type: (BaseSocket) -> None
        del self.fd_to_sock[sock.fileno()]
        sock.on_fail()
        self.noshot(sock)
        sock.close()

    def shutdown(self):
        """
        Forcibly close all sockets that this manages (calling their on_fail's),
        and close the object.

        This object is unusable after this call.
        """
        self.time_events = []
        for sock in list(six.itervalues(self.fd_to_sock)):
            sock.on_fail()
            sock.close()

        self.fd_to_sock = {}

    def oneshot(self, sock, delta, callback):
        """
        A socket registers a time callback
        :param sock: BaseSocket instance
        :param delta: "this seconds after now"
        :param callback: callable/0
        """
        if sock.fileno() in self.fd_to_sock:
            heapq.heappush(self.time_events, (monotonic() + delta,
                                              sock.fileno(),
                                              callback
                                              ))

    def activate(self, sock):  # type: (BaseSocket) -> None
        self.fd_to_sock[sock.fileno()] = sock

    @abstractmethod
    def register(self, sock,                    # type: socket.socket
                 on_read=lambda data: None,     # type: tp.Callable[[bytearray], None]
                 on_fail=lambda: None         # type: tp.Callable[[], None]
                 ):                     # type: () -> BaseSocket
        """
        This has to return a particular Socket instance, adapted to the needs of the listener.

        :param sock: a socket instance (as returned by socket module)
        :param on_read: callable(data) to be called with received data
        :param on_fail: callable() to be called when socket fails

        :return: a BaseSocket's subclass instance to use instead of this socket
        """



# coding=UTF-8
from __future__ import absolute_import, division, print_function

import six


# docs may be None


def as_unicode(clbl):
    def roll(*args, **kwargs):
        return six.text_type(clbl(*args, **kwargs))

    return roll


@as_unicode
def format_field_name(field):
    if field in (u'global', u'type'):
        field = field + '_'
    return field.replace('-', '_')



# coding=UTF-8
from __future__ import print_function, division, absolute_import

"""
That funny type, field-table...

A field-value is of form (value::any, type::unicode)

An array is of form [field-value1, field-value2, ...]

A table is of form ( (name1::bytes, fv1), (name2::bytes, fv2), ...)


NOTE: it's not buffers, it's memoryview all along
"""

import struct
import io
import six


def _tobuf(buf, pattern, *vals):  # type: (io.BytesIO, str, *tp.Any) -> int
    return buf.write(struct.pack(pattern, *vals))


def _tobufv(buf, value, pattern, *vals):  # type: (io.BytesIO, bytes, str, *tp.Any) -> None
    _tobuf(buf, pattern, *vals)
    buf.write(value)


def _frombuf(pattern, buf, offset):
    return struct.unpack_from(pattern, buf, offset)


def enframe_decimal(buf, v):  # convert decimal to bytes
    dps = 0
    for k in six.moves.xrange(20):
        k = v * (10 ** dps)
        if abs(k - int(k)) < 0.00001:  # epsilon
            return _tobuf(buf, '!BI', dps, k)

    raise ValueError('Could not convert %s to decimal', v)


def deframe_decimal(buf, offset):
    scale, val = _frombuf('!BI', buf, offset)
    return val / (10 ** scale), 5


def deframe_shortstr(buf, offset):  # -> value, bytes_eaten
    ln, = _frombuf('!B', buf, offset)
    return buf[offset + 1:offset + 1 + ln], 1 + ln


def enframe_shortstr(buf, value):
    _tobufv(buf, value, '!B', len(value))


def deframe_longstr(buf, offset):  # -> value, bytes_eaten
    ln, = _frombuf('!I', buf, offset)
    return buf[offset + 4:offset + 4 + ln], 4 + ln


def enframe_longstr(buf, value):
    _tobufv(buf, value, '!I', len(value))


def _c2none(buf, v):
    return None


FIELD_TYPES = {
    # length, struct, (option)to_bytes (callable(buffer, value)),
    #                 (option)from_bytes (callable(buffer, offset) ->
    #                                           value, bytes_consumed),
    #                 (option)get_len (callable(value) -> length in bytes)
    't': (1, '!?'),  # boolean
    'b': (1, '!b'),
    'B': (1, '!B'),
    'U': (2, '!H'),
    'u': (2, '!h'),
    'I': (4, '!I'),
    'i': (4, '!i'),
    'L': (8, '!Q'),
    'l': (8, '!q'),
    'f': (4, '!f'),
    'd': (8, '!d'),
    'D': (5, None, enframe_decimal, deframe_decimal),  # decimal-value
    's': (
        None, None, enframe_shortstr, deframe_shortstr,
        lambda val: len(val) + 1),
    # shortstr
    'S': (
        None, None, enframe_longstr, deframe_longstr,
        lambda val: len(val) + 4),
    # longstr
    'T': (8, '!Q'),
    'V': (0, None, _c2none, _c2none, 0),
    # rendered as None
}

if six.PY3:
    chrpy3 = chr
else:
    chrpy3 = lambda x: x


def enframe_field_value(buf, fv):
    value, type = fv
    buf.write(type.encode('utf8'))

    opt = FIELD_TYPES[type]

    if opt[1] is not None:
        _tobuf(buf, opt[1], value)
    else:
        opt[2](buf, value)


def deframe_field_value(buf, offset):  # -> (value, type), bytes_consumed
    start_offset = offset
    field_type = chrpy3(buf[offset])
    offset += 1

    if field_type not in FIELD_TYPES.keys():
        raise ValueError('Unknown field type %s!', (repr(field_type),))

    opt = FIELD_TYPES[field_type]

    if opt[1] is not None:
        field_val, = struct.unpack_from(FIELD_TYPES[field_type][1], buf,
                                        offset)
        offset += opt[0]
    else:
        field_val, delta = opt[3](buf, offset)
        offset += delta

    return (field_val, field_type), offset - start_offset


def deframe_array(buf, offset):
    start_offset = offset
    ln, = _frombuf('!I', buf, offset)
    offset += 4

    values = []
    while offset < (start_offset + 1 + ln):
        v, t, delta = deframe_field_value(buf, offset)
        offset += delta
        values.append((v, t))

    if offset != start_offset + 4 + ln:
        raise ValueError(
            'Array longer than expected, took %s, expected %s bytes',
            (offset - (start_offset + ln + 4), ln + 4))

    return values, ln + 4


def enframe_array(buf, array):
    _tobuf(buf, '!I', frame_array_size(array) - 4)
    for fv in array:
        enframe_field_value(buf, fv)


def enframe_table(buf, table):  # type (tp.BinaryIO, table) -> None
    """
    Write AMQP table to buffer

    :param buf: target buffer to write to
    :param table: table to write
    """
    _tobuf(buf, '!I', frame_table_size(table) - 4)

    for name, fv in table:
        _tobufv(buf, name, '!B', len(name))
        enframe_field_value(buf, fv)


def deframe_table(buf, start_offset):  # -> (table, bytes_consumed)
    """:return: tuple (table, bytes consumed)"""
    offset = start_offset
    table_length, = struct.unpack_from('!L', buf, start_offset)
    offset += 4

    # we will check if it's really so.
    fields = []

    while offset < (start_offset + table_length + 4):
        field_name, ln = deframe_shortstr(buf, offset)
        offset += ln
        fv, delta = deframe_field_value(buf, offset)
        offset += delta
        fields.append((field_name.tobytes(), fv))

    if offset > (start_offset + table_length + 4):
        raise ValueError(
            'Table turned out longer than expected! Found %s bytes expected %s',
            (offset - start_offset, table_length))

    return fields, table_length + 4


def frame_field_value_size(fv):
    v, t = fv
    if FIELD_TYPES[t][0] is None:
        return FIELD_TYPES[t][4](v) + 1
    else:
        return FIELD_TYPES[t][0] + 1


def frame_array_size(array):
    return 4 + sum(frame_field_value_size(fv) for fv in array)


def frame_table_size(table):
    """
    :return: length of table representation, in bytes, INCLUDING length
     header"""
    return 4 + sum(1 + len(k) + frame_field_value_size(fv) for k, fv in table)


FIELD_TYPES['A'] = (None, None, enframe_array, deframe_array, frame_array_size)
FIELD_TYPES['F'] = (None, None, enframe_table, deframe_table, frame_table_size)



# coding=UTF-8
from __future__ import absolute_import, division, print_function

import logging
import select
import socket

import six

from coolamqp.uplink.listener.socket import SocketFailed, BaseSocket
from coolamqp.uplink.listener.base_listener import BaseListener

logger = logging.getLogger(__name__)


class SelectListener(BaseListener):
    """
    A listener using select
    """

    def wait(self, timeout=0.5):
        rds_and_exs = []        # waiting both for read and for exception
        wrs = []                # waiting for write
        for sock in six.itervalues(self.fd_to_sock):
            rds_and_exs.append(sock)
            if sock.wants_to_send_data():
                wrs.append(sock)

        self.do_timer_events()

        try:
            rds, wrs, exs = select.select(rds_and_exs, wrs, rds_and_exs, timeout)
        except (select.error, socket.error, IOError):
            for sock in rds_and_exs:
                try:
                    select.select([sock], [], [], timeout=0)
                except (select.error, socket.error, IOError):
                    self.close_socket(sock)
                    return
            else:
                return

        for sock_rd in rds:
            try:
                sock_rd.on_read()
            except SocketFailed:
                return self.close_socket(sock_rd)

        for sock_wr in wrs:
            try:
                sock_wr.on_write()
            except SocketFailed:
                return self.close_socket(sock_wr)

        for sock_ex in exs:
            try:
                sock_rd.on_read()
            except SocketFailed:
                return self.close_socket(sock_ex)

    def register(self, sock, on_read=lambda data: None,
                 on_fail=lambda: None):
        """
        Add a socket to be listened for by the loop.

        :param sock: a socket instance (as returned by socket module)
        :param on_read: callable(data) to be called with received data
        :param on_fail: callable() to be called when socket fails

        :return: a BaseSocket instance to use instead of this socket
        """
        return BaseSocket(sock, on_read, on_fail=on_fail, listener=self)



# coding=UTF-8
"""
Cluster will emit Events.

They mean that something, like, happened.
"""
from __future__ import print_function, absolute_import, division

import logging

from coolamqp.objects import ReceivedMessage

logger = logging.getLogger(__name__)


class Event(object):
    """
    An event emitted by Cluster
    """

    __slots__ = ()


class NothingMuch(Event):
    """Nothing happened :D"""

    __slots__ = ()


class ConnectionLost(Event):
    """
    We have lost the connection.

    NOTE that we don't have a ConnectionUp, since re-establishing a connection
    is a complicated process. Broker might not have recognized the failure,
    and some queues will be blocked, some might be ok, and the state
    might be just a bit noisy.

    Please examine your Consumer's .state's to check whether link was regained
    """

    __slots__ = ()


class MessageReceived(ReceivedMessage, Event):
    """
    Something that works as an ersatz ReceivedMessage, but is an event
    """
    __slots__ = ()

    def __init__(self, msg):  # type: (ReceivedMessage) -> None
        """:type msg: ReceivedMessage"""
        ReceivedMessage.__init__(self, msg.body, msg.exchange_name,
                                 msg.routing_key,
                                 properties=msg.properties,
                                 delivery_tag=msg.delivery_tag,
                                 ack=msg.ack, nack=msg.nack)



# coding=UTF-8
from __future__ import absolute_import, division, print_function

"""
Provides reactors that can authenticate an AQMP session
"""
import six
import typing as tp
import copy
import logging
from coolamqp.framing.definitions import ConnectionStart, ConnectionStartOk, \
    ConnectionTune, ConnectionTuneOk, ConnectionOpen, ConnectionOpenOk
from coolamqp.framing.frames import AMQPMethodFrame
from coolamqp.uplink.connection.states import ST_ONLINE
from coolamqp.uplink.heartbeat import Heartbeater
from coolamqp import __version__

PUBLISHER_CONFIRMS = b'publisher_confirms'
CONSUMER_CANCEL_NOTIFY = b'consumer_cancel_notify'
CONNECTION_BLOCKED = b'connection.blocked'

SUPPORTED_EXTENSIONS = [
    PUBLISHER_CONFIRMS,
    CONSUMER_CANCEL_NOTIFY,     # half assed support - we just .cancel the consumer, see #12
    CONNECTION_BLOCKED
]

CLIENT_DATA = [
    # because RabbitMQ is some kind of a fascist and does not allow
    # these fields to be of type short-string
    (b'product', (b'CoolAMQP', 'S')),
    (b'version', (__version__.encode('utf8'), 'S')),
    (b'copyright', (b'Copyright (C) 2016-2020 SMOK sp. z o.o.', 'S')),
    (
        b'information', (
            b'Licensed under the MIT License.\nSee https://github.com/smok-serwis/coolamqp for details',
            'S')),
    (b'capabilities',
     ([(capa, (True, 't')) for capa in SUPPORTED_EXTENSIONS], 'F')),
]

WATCHDOG_TIMEOUT = 10

logger = logging.getLogger(__name__)


class Handshaker(object):
    """
    Object that given a connection rolls the handshake.
    """

    def __init__(self, connection,  # type: coolamqp.uplink.connection.Connection
                 node_definition,  # type: coolamqp.objects.NodeDefinition
                 on_success,  # type: tp.Callable[[], None]
                 extra_properties=None  # type: tp.Dict[bytes, tp.Tuple[tp.Any, str]]
                 ):
        """
        :param connection: Connection instance to use
        :type node_definition: NodeDefinition
        :param on_success: callable/0, on success
        """
        self.connection = connection
        self.login = node_definition.user.encode('utf8')
        self.password = node_definition.password.encode('utf8')
        self.virtual_host = node_definition.virtual_host.encode('utf8')
        self.heartbeat = node_definition.heartbeat or 0
        self.connection.watch_for_method(0, ConnectionStart,
                                         self.on_connection_start)

        # Callbacks
        self.on_success = on_success
        self.EXTRA_PROPERTIES = extra_properties or []

    # Called by internal setup
    def on_watchdog(self):
        """
        Called WATCHDOG_TIMEOUT seconds after setup begins

        If we are not ST_ONLINE after that much, something is wrong and pwn this connection.
        """
        # Not connected in 20 seconds - abort
        if self.connection.state != ST_ONLINE:
            # closing the connection this way will get to Connection by channels of ListenerThread
            self.connection.send(None)

    def on_connection_start(self, payload  # type: coolamqp.framing.base.AMQPPayload
                            ):

        sasl_mechanisms = payload.mechanisms.tobytes().split(b' ')
        locale_supported = payload.locales.tobytes().split(b' ')

        # Select a mechanism
        if b'PLAIN' not in sasl_mechanisms:
            raise ValueError('Server does not support PLAIN')

        # Select capabilities
        server_props = dict(payload.server_properties)
        if b'capabilities' in server_props:
            for label, fv in server_props[b'capabilities'][0]:
                if label in SUPPORTED_EXTENSIONS:
                    if fv[0]:
                        self.connection.extensions.append(label)

        self.connection.watchdog(WATCHDOG_TIMEOUT, self.on_watchdog)
        self.connection.watch_for_method(0, ConnectionTune,
                                         self.on_connection_tune)

        CLIENT_DATA_c = copy.copy(CLIENT_DATA)
        CLIENT_DATA_c.extend(self.EXTRA_PROPERTIES)
        self.connection.send([
            AMQPMethodFrame(0,
                            ConnectionStartOk(CLIENT_DATA_c, b'PLAIN',
                                              b'\x00' + self.login + b'\x00' + self.password,
                                              locale_supported[0]
                                              ))
        ])

    def on_connection_tune(self, payload  # type: coolamqp.framing.base.AMQPPayload
                           ):
        self.connection.frame_max = payload.frame_max
        self.connection.heartbeat = min(payload.heartbeat, self.heartbeat)
        self.connection.free_channels.extend(six.moves.xrange(1, (
            65535 if payload.channel_max == 0 else payload.channel_max) + 1))

        self.connection.watch_for_method(0, ConnectionOpenOk,
                                         self.on_connection_open_ok)
        self.connection.send([
            AMQPMethodFrame(0, ConnectionTuneOk(payload.channel_max,
                                                payload.frame_max,
                                                self.connection.heartbeat)),
            AMQPMethodFrame(0, ConnectionOpen(self.virtual_host))
        ])

        # Install heartbeat handlers NOW, if necessary
        if self.connection.heartbeat > 0:
            Heartbeater(self.connection, self.connection.heartbeat)

    def on_connection_open_ok(self, payload  # type: coolamqp.framing.base.AMQPPayload
                              ):
        self.on_success()



# coding=UTF-8
from __future__ import absolute_import, division, print_function

import logging

from coolamqp.framing.base import AMQPMethodPayload
from coolamqp.framing.frames import AMQPMethodFrame, AMQPHeaderFrame, \
    AMQPBodyFrame

logger = logging.getLogger(__name__)


class Watch(object):
    """
    A watch is placed per-channel, to listen for a particular frame.

    Multiple watches can be registered to listen for a single frame. All watches will be fired then.
    """

    class CancelMe(Exception):
        """To be raised in a watch if it wants to be cancelled"""

    def __init__(self, channel, oneshot):
        """
        :param channel: Channel to listen to.
            all channels if None is passed
        :param oneshot: Is destroyed after triggering?
        """
        self.channel = channel
        self.oneshot = oneshot
        self.cancelled = False

    def is_triggered_by(self, frame):
        """
        Does frame trigger this watch?
        Run callable if it does.
        :param frame: AMQPFrame instance
        :return: bool
        """
        raise Exception('Abstract method')

    def failed(self):
        """
        This watch will process things no more, because underlying
        link has failed
        """

    def cancel(self):
        """
        Called by watch's user. This watch will not receive events anymore
        (whether about frame or fail), and it will be discarded upon next iteration.
        """
        self.cancelled = True


class AnyWatch(Watch):
    """
    Watch that listens for any frame.

    It does not listen for failures.

    Used because heartbeating is implemented improperly EVERYWHERE
    (ie. you might not get a heartbeat when connection is so loaded it just can't get it in time,
    due to loads and loads of message exchanging).

    Eg. RabbitMQ will happily disconnect you if you don't, but it can get lax with heartbeats
    as it wants.
    """

    def __init__(self, callable):
        super(AnyWatch, self).__init__(None, False)
        self.callable = callable

    def is_triggered_by(self, frame):
        try:
            self.callable(frame)
        except Watch.CancelMe:
            self.cancel()
        return True


class FailWatch(Watch):
    """
    A special kind of watch that fires when connection has died
    """

    def __init__(self, callable):
        super(FailWatch, self).__init__(None, True)
        self.callable = callable

    def is_triggered_by(self, frame):
        return False

    def failed(self):
        """Connection failed!"""
        self.callable()


class HeaderOrBodyWatch(Watch):
    """
    A multi-shot watch listening for AMQP header or body frames
    """

    def __init__(self, channel, callable):
        Watch.__init__(self, channel, False)
        self.callable = callable

    def is_triggered_by(self, frame):
        if not (isinstance(frame, (AMQPHeaderFrame, AMQPBodyFrame))):
            return False
        try:
            self.callable(frame)
        except Watch.CancelMe:
            self.cancel()
        return True


class MethodWatch(Watch):
    """
    One-shot watch listening for methods.
    """

    def __init__(self, channel, method_or_methods, callable, on_end=None):
        """
        :param method_or_methods: class, or list of AMQPMethodPayload classes
        :param callable: callable(AMQPMethodPayload instance)
        :param on_end: callable/0 on link dying
        """
        Watch.__init__(self, channel, True)
        self.callable = callable
        if isinstance(method_or_methods, (list, tuple)):
            self.methods = tuple(method_or_methods)
        elif issubclass(method_or_methods, AMQPMethodPayload):
            self.methods = (method_or_methods,)
        self.on_end = on_end

    def __repr__(self):
        return '<MethodWatch %s, %s, %s, on_end=%s>' % (self.channel, self.methods, self.callable, self.on_end)

    def failed(self):
        if self.on_end is not None:
            self.on_end()

    def is_triggered_by(self, frame):

        if not isinstance(frame, AMQPMethodFrame):
            return False

        if isinstance(frame.payload, self.methods):
            try:
                self.callable(frame.payload)
            except Watch.CancelMe:
                self.cancel()
            return True
        return False



import logging

from coolamqp.framing.frames import AMQPMethodFrame


class BaseFrameTracer(object):
    """An abstract do-nothing frame tracer"""

    def on_frame(self, timestamp, frame, direction):
        """
        Called by AMQP upon receiving a frame information

        :param timestamp: timestamp
        :type timestamp: float
        :param frame: frame that is sent or received
        :type frame: :class:`coolamqp.framing.base.AMQPFrame`
        :param direction: either 'to_client' or 'to_server'
        :type direction: str
        """


class LoggingFrameTracer(BaseFrameTracer):
    """
    A frame tracer that outputs each frame to log

    :param logger: the logger to log onto
    :param log_level: the level of logging to log with
    """
    def __init__(self, logger, log_level=logging.WARNING):
        self.logger = logger
        self.log_level = log_level

    def on_frame(self, timestamp, frame, direction):
        if direction == 'to_client':
            if isinstance(frame, AMQPMethodFrame):
                self.logger.log(self.log_level, 'RECEIVED METHOD %s', frame.payload)
            else:
                self.logger.log(self.log_level, 'RECEIVED %s type %s', frame, type(frame))
        else:
            if isinstance(frame, AMQPMethodFrame):
                self.logger.log(self.log_level, 'SENT METHOD %s', frame.payload)
            else:
                self.logger.log(self.log_level, 'SENT %s type %s', frame, type(frame))


class HoldingFrameTracer(BaseFrameTracer):
    """
    A frame tracer that holds the frames in memory

    :ivar frames: a list of tuple (direction:str (either 'to_client' or 'to_server'),
                              timestamp::float,
                              frame:: :class:`~coolamqp.framing.base.AMQPFrame`)
    """
    def __init__(self):
        self.frames = []

    def on_frame(self, timestamp, frame, direction):
        self.frames.append((direction, timestamp, frame))

    def clear(self):
        """Clear internal frame list"""
        self.frames = []



#!/usr/bin/env python
# coding=UTF-8
from setuptools import setup, find_packages

from coolamqp import __version__

setup(keywords=['amqp', 'rabbitmq', 'client', 'network', 'ha', 'high availability'],
      version=__version__,
      packages=find_packages(include=['coolamqp', 'coolamqp.*']),
      install_requires=['six'],
      # per coverage version for codeclimate-reporter
      tests_require=["nose2", "coverage", "nose2[coverage_plugin]"],
      test_suite='nose2.collector.collector',
      python_requires='!=2.7.*,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*',
      extras_require={
          ':python_version == "2.7"': ['futures', 'typing', 'monotonic'],
          'prctl': ['prctl'],
          'opentracing': ['opentracing'],
          'gevent': ['gevent']
      }
      )



# coding=UTF-8
"""
Base class for consumer or publisher with the capabiility to
set up and tear down channels
"""
from __future__ import print_function, absolute_import, division

import logging
import typing as tp

from coolamqp.framing.definitions import ChannelOpen, ChannelOpenOk, \
    ChannelClose, ChannelCloseOk, BasicCancel, \
    BasicCancelOk
from coolamqp.framing.frames import AMQPMethodFrame

ST_OFFLINE = 0  # Consumer is *not* consuming, no setup attempts are being made
ST_SYNCING = 1  # A process targeted at consuming has been started
ST_ONLINE = 2  # Consumer is declared all right

logger = logging.getLogger(__name__)


class Attache(object):
    """
    Something that can be attached to connection.
    """
    __slots__ = ('cancelled', 'state', 'connection')

    def __init__(self):
        self.cancelled = False  #: public, if this is True, it won't be attached to next connection
        self.state = ST_OFFLINE
        self.connection = None

    def attach(self, connection):  # type (coolamqp.uplink.connection.Connection) -> None
        """
        Attach to a connection.

        :param connection: Connection instance of any state
        """
        assert self.connection is None
        assert connection.state != ST_OFFLINE
        self.connection = connection


class Channeler(Attache):
    """
    A base class for Consumer/Publisher implementing link set up and tear down.

    A channeler can be essentially in 4 states:
    - ST_OFFLINE (.channel is None): channel is closed, object is unusable. Requires an attach() a connection
                                     that is being established, or open, or whatever. Connection will notify
                                     this channeler that it's open.
    - ST_SYNCING: channeler is opening a channel/doing some other things related to it's setup.
                  it's going to be ST_ONLINE soon, or go back to ST_OFFLINE.
                  It has, for sure, acquired a channel number.
    - ST_ONLINE:  channeler is operational. It has a channel number and has declared everything
                  it needs to.

                  on_operational(True) will be called when a transition is made TO this state.
                  on_operational(False) will be called when a transition is made FROM this state.

    - ST_OFFLINE (.channel is not None): channeler is undergoing a close. It has not yet torn down the channel,
                                         but ordering it to do anything is pointless, because it will not get done
                                         until attach() with new connection is called.
    """
    __slots__ = ('channel_id',)

    def __init__(self):
        """
        [EXTEND ME!]
        """
        super(Channeler, self).__init__()
        self.channel_id = None  # channel obtained from Connection

    def attach(self, connection):  # type (coolamqp.uplink.connection.Connection) -> None
        """
        Attach this object to a live Connection.

        :param connection: Connection instance to use
        """
        super(Channeler, self).attach(connection)
        assert self.connection is not None
        connection.call_on_connected(self.on_uplink_established)

    # ------- event handlers

    def on_operational(self, operational):  # type: (bool) -> None
        """
        [EXTEND ME] Called by internal methods (on_*) when channel has achieved (or lost) operational status.

        If this is called with operational=True, then for sure it will be called with operational=False.

        This will, therefore, get called an even number of times.

        Called by Channeler, when:
            - Channeler.on_close gets called and state is ST_ONLINE
                on_close registers ChannelClose, ChannelCloseOk, BasicCancel

        :param operational: True if channel has just become operational, False if it has just become useless.
        """

    def on_close(self, payload=None):
        # type (tp.Optional[coolamqp.framing.base.AMQPMethodPayload) -> None
        """
        [EXTEND ME] Handler for channeler destruction.

        Channeler registers this for:
            (None - socket dead)
            (BasicCancel, BasicCancelOk, ChannelCloseOk, ChannelClose)

        This method provides to send a response for ChannelClose

        This handles following situations:
        - payload is None: this means that connection has gone down hard, so our Connection object is
                           probably very dead. Transition to ST_OFFLINE (.channel is None)
        - payload is a ChannelClose: this means that a channel exception has occurred. Dispatch a ChannelCloseOk,
                                     attempt to log an exception, transition to ST_OFFLINE (.channel is None)
        - payload is a ChannelCloseOk: this means that it was us who attempted to close the channel. Return the channel
                                       to free pool, transition to ST_OFFLINE (.channel is None)

        If you need to handle something else, extend this. Take care that this DOES NOT HANDLE errors that happen
        while state is ST_SYNCING. You can expect this to handle a full channel close, therefore releasing all
        resources, so it mostly will do *the right thing*.

        If you need to do something else than just close a channel, please extend or modify as necessary.

        WARNING: THIS WILL GET CALLED TWICE.
            Once on ChannelClose - if so,
            Second with None - because socket dies.

            Be prepared!

        """
        if self.connection is None:
            # teardown already done
            return

        if self.state == ST_ONLINE:  # The channel has just lost operationality! Inform others ASAP.
            self.on_operational(False)
        self.state = ST_OFFLINE

        if not isinstance(payload, (ChannelClose, ChannelCloseOk)) and (
                payload is not None):
            # I do not know how to handle that!
            return

        if isinstance(payload, ChannelClose):
            # it would still be good to reply with channel.close-ok
            self.method(ChannelCloseOk())

        if payload is not None:
            assert self.channel_id is not None
            self.connection.free_channels.append(self.channel_id)
            # it's just dead don't bother with returning port
            # at this point, this channel might still have some watches,
            # especially if it was interrupted unexpectedly.
            # clean up!
            self.connection.unwatch_all(self.channel_id)

        self.connection = None
        self.channel_id = None

        if isinstance(payload, ChannelClose):
            logger.debug('Channel closed: %s %s', payload.reply_code,
                         payload.reply_text.tobytes())

    def methods(self, payloads):
        # type: (tp.Iterable[coolamqp.framing.base.AMQPMethodPayload]) -> None
        """
        Syntactic sugar for

            for payload in paylods:
                self.method(payload)

        But moar performant.
        """
        if self.channel_id is None:
            return  # advanced teardown xD

        frames = [AMQPMethodFrame(self.channel_id, payload) for payload in
                  payloads]
        self.connection.send(frames)

    def method(self, payload):
        # type: (tp.Iterable[coolamqp.framing.base.AMQPMethodPayload]) -> None
        """
        Syntactic sugar for:

            self.connection.send([AMQPMethodFrame(self.channel_id, payload)])
        """
        self.methods([payload])

    def watch_for_method(self, method, callback, on_fail=None):  # type: () -> MethodWatch
        """
        Syntactic sugar for

        >>> self.connection.watch_for_method(self.channel_id, method, callback, on_fail=on_fail)
        """
        assert self.channel_id is not None
        return self.connection.watch_for_method(self.channel_id, method, callback, on_fail=on_fail)

    def method_and_watch(self, method_payload,      # type: coolamqp.framing.base.AMQPMethodPayload,
                         method_classes_to_watch,   # type: tp.Iterable[tp.Type[AMQPMethodPayload]]
                         callable                   # type: tp.Callable[[AMQPMethodPayload], None]
                         ):     # type: () -> None
        """
        Syntactic sugar for

        >>> self.connection.method_and_watch(self.channel_id,
        >>>                                  method_payload,
        >>>                                  method_classes_to_watch,
        >>>                                  callable)
        """
        assert self.channel_id is not None
        self.connection.method_and_watch(self.channel_id, method_payload,
                                         method_classes_to_watch, callable)

    def on_setup(self, payload):  # type: (coolamqp.framing.base.AMQPMethodPayload) -> None
        """
        [OVERRIDE ME!] Called with a method frame that signifies a part of setup.

        You must be prepared to handle at least a payload of ChannelOpenOk

        :param payload: AMQP method frame payload
        """
        raise Exception('Abstract method - override me!')

    def register_on_close_watch(self):
        """
        Register a watch for on_close.

        Since on_close is a one-shot, it will expire upon calling.

        To be called by on_close, when it needs to be notified just one more time.
        """
        self.connection.watch_for_method(self.channel_id, (
            ChannelClose, ChannelCloseOk, BasicCancel, BasicCancelOk),
                                         self.on_close,
                                         on_fail=self.on_close)

    def on_uplink_established(self):
        """Called by connection. Connection reports being ready to do things."""
        assert self.connection is not None
        assert self.connection.state == ST_ONLINE, repr(self)
        self.state = ST_SYNCING
        self.channel_id = self.connection.free_channels.pop()
        self.register_on_close_watch()

        self.connection.method_and_watch(
            self.channel_id,
            ChannelOpen(),
            ChannelOpenOk,
            self.on_setup
        )



# coding=UTF-8
from __future__ import absolute_import, division, print_function

"""
Attaches are components that attach to an coolamqp.uplink.Connection and perform some duties
These duties almost always require allocating a channel. A base class - Channeler - is provided to faciliate that.
The attache becomes then responsible for closing this channel.

Attache should also register at least one on_fail watch, so it can handle things if they go south.

Multiple attaches can be "abstracted" as single one via AttacheGroup (which is also an Attache)

EVERYTHING HERE IS CALLED BY LISTENER THREAD UNLESS STATED OTHERWISE.
"""

from coolamqp.attaches.consumer import Consumer, BodyReceiveMode
from coolamqp.attaches.publisher import Publisher
from coolamqp.attaches.agroup import AttacheGroup
from coolamqp.attaches.declarer import Declarer



# coding=UTF-8
"""
Core objects used in CoolAMQP
"""
import logging
import typing as tp
import uuid

import six

from coolamqp.framing.base import AMQPFrame
from coolamqp.framing.definitions import \
    BasicContentPropertyList as MessageProperties

logger = logging.getLogger(__name__)

EMPTY_PROPERTIES = MessageProperties()


def toutf8(q):
    if isinstance(q, six.binary_type):
        q = q.decode('utf8')
    return q


def tobytes(q):
    if isinstance(q, six.text_type):
        q = q.encode('utf8')
    return q


class Callable(object):
    """
    Add a bunch of callables to one list, and just invoke'm.
    INTERNAL USE ONLY
    #todo not thread safe
    """
    __slots__ = ('callables', 'oneshots')

    def __init__(self, oneshots=False):
        """:param oneshots: if True, callables will be called and discarded"""
        self.callables = []
        self.oneshots = oneshots

    def add(self, callable):
        self.callables.append(callable)

    def __call__(self, *args, **kwargs):
        for callable in self.callables:
            callable(*args, **kwargs)
        if self.oneshots:
            self.callables = []


class Message(object):
    """
    An AMQP message. Has a binary body, and some properties.

    Properties is a highly regularized class - see
    coolamqp.framing.definitions.BasicContentPropertyList
    for a list of possible properties.

    :param body: stream of octets
    :type body: anything with a buffer interface
    :param properties: AMQP properties to be sent along.
                       default is 'no properties at all'
                       You can pass a dict - it will be passed to
                       MessageProperties,
                       but it's slow - don't do that.
    :type properties: MessageProperties instance, None or a dict (SLOW!)

    """
    __slots__ = ('body', 'properties')

    Properties = MessageProperties  # an alias for easier use

    def __init__(self, body,         # type: bytes
                 properties=None     # type: tp.Optional[MessageProperties]
                 ):
        """
        Create a Message object.

        Please take care with passing empty bodies, as py-amqp has some
        failure on it.
        """
        if isinstance(body, six.text_type):
            raise TypeError(u'body cannot be a text type!')

        self.body = body

        if isinstance(properties, dict):
            self.properties = MessageProperties(**properties)
        elif properties is None:
            self.properties = EMPTY_PROPERTIES
        else:
            self.properties = properties


def LAMBDA_NONE():
    pass


class ReceivedMessage(Message):
    """
    A message that was received from the AMQP broker.

    It additionally has an exchange name, routing key used, it's delivery tag,
    and methods for ack() or nack().

    Note that if the consumer that generated this message was no_ack, .ack()
    and .nack() are no-ops.
    """
    __slots__ = ('delivery_tag', 'exchange_name', 'routing_key', '_ack', '_nack',
                 'acked')

    def __init__(self, body,  # type: tp.Union[str, bytes, bytearray, tp.List[memoryview]]
                 exchange_name,  # type: memoryview
                 routing_key,  # type: memoryview
                 properties=None,
                 delivery_tag=None,  # type: int
                 ack=None,  # type: tp.Callable[[], None]
                 nack=None  # type: tp.Callable[[], None]
                 ):
        """
        :param body: message body. A stream of octets.
        :type body: str (py2) or bytes (py3) or a list of memoryviews, if
            particular disabled-by-default option is turned on, or a single memoryview
        :param exchange_name: name of exchange this message was submitted to
        :param routing_key: routing key with which this message was sent
        :param properties: a suitable BasicContentPropertyList subinstance.
                           be prepared that value of properties that are
                           strings will be memoryviews
        :param delivery_tag: delivery tag assigned by AMQP broker to confirm
            this message
        :param ack: a callable to call when you want to ack (via basic.ack)
            this message. None if received by the no-ack mechanism
        :param nack: a callable to call when you want to nack
            (via basic.reject) this message. None if received by the no-ack
             mechanism
        """
        Message.__init__(self, body, properties=properties)

        self.delivery_tag = delivery_tag
        self.exchange_name = exchange_name
        self.routing_key = routing_key
        self.acked = False
        self._ack = ack or LAMBDA_NONE
        self._nack = nack or LAMBDA_NONE

    def ack(self):
        """
        Acknowledge reception of this message.

        This is a no-op if a Consumer was called with no_ack=True.

        If called after an ack() or nack() was called, this will be a no-op.
        """
        if self.acked:
            return
        self._ack()
        self.acked = True

    def nack(self):
        """
        Negatively acknowledge reception of this message.

        This is a no-op if a Consumer was called with no_ack=True. If no_ack was False,
        the message will be requeued and redelivered by the broker

        If called after an ack() or nack() was called, this will be a no-op.
        """
        if self.acked:
            return
        self._nack()
        self.acked = True


class Exchange(object):
    """
    This represents an Exchange used in AMQP.
    This is hashable.
    """
    __slots__ = ('name', 'type', 'durable', 'auto_delete')

    direct = None  # the direct exchange

    def __init__(self, name=u'',  # type: tp.Union[str, bytes]
                 type=b'direct',  # type: tp.Union[str, bytes]
                 durable=True,  # type: bool
                 auto_delete=False  # type: bool
                 ):
        """
        :type name: unicode is preferred, binary type will get decoded to
             unicode with utf8
        """
        self.name = toutf8(name)  # must be unicode
        self.type = tobytes(type)  # must be bytes
        self.durable = durable
        self.auto_delete = auto_delete

        assert isinstance(self.name, six.text_type)
        assert isinstance(self.type, six.binary_type)

    def __repr__(self):  # type: () -> str
        return u'Exchange(%s, %s, %s, %s)' % (
            repr(self.name), repr(self.type), repr(self.durable),
            repr(self.auto_delete))

    def __hash__(self):  # type: () -> int
        return self.name.__hash__()

    def __eq__(self, other):  # type: (Exchange) -> bool
        return (self.name == other.name) and (type(self) == type(other))


Exchange.direct = Exchange()


class Queue(object):
    """
    This object represents a Queue that applications consume from or publish to.
    """
    __slots__ = ('name', 'durable', 'exchange', 'auto_delete', 'exclusive',
                 'anonymous', 'consumer_tag')

    def __init__(self, name=b'',  # type: tp.Union[str, bytes]
                 durable=False,  # type: bool
                 exchange=None,  # type: tp.Optional[Exchange]
                 exclusive=False,  # type: bool
                 auto_delete=False  # type: bool
                 ):
        """
        Create a queue definition.

        :param name: name of the queue.
            Take special care if this is empty. If empty, this will be filled-in by the broker
            upon declaration. If a disconnect happens, and connection to other node is
            reestablished, this name will CHANGE AGAIN, and be reflected in this object.
            This change will be done before CoolAMQP signals reconnection.
        :param durable: Is the queue durable?
        :param exchange: Exchange for this queue to bind to. None for no binding.
        :param exclusive: Is this queue exclusive?
        :param auto_delete: Is this queue auto_delete ?
        """
        self.name = tobytes(name)  #: public, must be bytes
        # if name is '', this will be filled in with broker-generated name upon declaration
        self.durable = durable
        self.exchange = exchange
        self.auto_delete = auto_delete
        self.exclusive = exclusive

        self.anonymous = len(
            self.name) == 0  # if this queue is anonymous, it must be regenerated upon reconnect

        self.consumer_tag = self.name if not self.anonymous else uuid.uuid4().hex.encode(
            'utf8')  # bytes, consumer tag to use in AMQP comms

        assert isinstance(self.name, six.binary_type)
        assert isinstance(self.consumer_tag, six.binary_type)

    def __eq__(self, other):
        return (self.name == other.name) and (type(self) == type(other))

    def __hash__(self):
        return hash(self.name)


class QueueBind(object):
    """An order to be declared which binds a given queue to an exchange"""
    def __init__(self, queue,   # type: tp.Union[Queue, bytes, unicode]
                 exchange,      # type: tp.Union[Exchange, bytes, unicode]
                 routing_key    # type: tp.Union[bytes, unicode]
                 ):
        if isinstance(queue, Queue):
            queue = queue.name
        self.queue = tobytes(queue)        # type: bytes
        if isinstance(exchange, Exchange):
            exchange = exchange.name
        self.exchange = tobytes(exchange)   # type: bytes
        self.routing_key = tobytes(routing_key)     # type: bytes

    def __eq__(self, other):
        return self.queue == other.queue and self.exchange == other.exchange and self.routing_key == other.routing_key

    def __hash__(self):
        return hash(self.queue) ^ hash(self.exchange) ^ hash(self.routing_key)


class NodeDefinition(object):
    """
    Definition of a reachable AMQP node.

    This object is hashable.

    >>> a = NodeDefinition(host='192.168.0.1', user='admin', password='password',
    >>>                   virtual_host='vhost')

    or

    >>> a = NodeDefinition('192.168.0.1', 'admin', 'password')

    or

    >>> a = NodeDefinition('amqp://user:password@host/virtual_host')

    or

    >>> a = NodeDefinition('amqp://user:password@host:port/virtual_host', hearbeat=20)

    AMQP connection string may be either bytes or str/unicode


    Additional keyword parameters that can be specified:
        heartbeat - heartbeat interval in seconds
        port - TCP port to use. Default is 5672

    :raise ValueError: invalid parameters
    """

    def __init__(self, *args, **kwargs):
        self.heartbeat = kwargs.pop('heartbeat', None)
        self.port = kwargs.pop('port', 5672)

        if len(kwargs) > 0:
            # Prepare arguments for amqp.connection.Connection
            self.host = kwargs['host']
            self.user = kwargs['user']
            self.password = kwargs['password']
            self.virtual_host = kwargs.get('virtual_host', '/')
        elif len(args) == 3:
            self.host, self.user, self.password = args
            self.virtual_host = '/'
        elif len(args) == 4:
            self.host, self.user, self.password, self.virtual_host = args
        elif len(args) == 1 and isinstance(args[0],
                                           (six.text_type, six.binary_type)):
            connstr = args[0].decode('utf8') if isinstance(args[0],
                                                           six.binary_type) else \
                args[0]
            # AMQP connstring
            if not connstr.startswith(u'amqp://'):
                raise ValueError(u'should begin with amqp://')

            connstr = connstr.replace(u'amqp://', u'')
            self.user, connstr = connstr.split(u':', 1)
            self.password, connstr = connstr.split(u'@', 1)
            self.host, self.virtual_host = connstr.split(u'/', 1)

            if len(self.virtual_host) == 0:
                # empty virtual host is /
                self.virtual_host = u'/'

            if u':' in self.host:
                host, port = self.host.split(u':', 1)
                self.port = int(port)
                # else get that port from kwargs
        else:
            raise ValueError(u'What did you exactly pass?')

    def __str__(self):  # type: () -> str
        return six.text_type(
            'amqp://%s:%s@%s/%s' % (self.host, self.port, self.user, self.virtual_host))
